/*++
/* NAME
/*	cleanup_bounce 3
/* SUMMARY
/*	process extracted segment
/* SYNOPSIS
/*	#include "cleanup.h"
/*
/*	void	cleanup_bounce(state)
/*	CLEANUP_STATE *state;
/* DESCRIPTION
/*	cleanup_bounce() returns undeliverable mail to the sender,
/*	on request by client programs that cannot handle such
/*	problems themselves.
/*
/*	Upon successful completion, all error flags are reset.
/*	Otherwise, the CLEANUP_STAT_WRITE error flag is raised.
/*
/*	Arguments:
/* .IP state
/*	Queue file and message processing state. This state is
/*	updated as records are processed and as errors happen.
/* LICENSE
/* .ad
/* .fi
/*	The Secure Mailer license must be distributed with this software.
/* AUTHOR(S)
/*	Wietse Venema
/*	IBM T.J. Watson Research
/*	P.O. Box 704
/*	Yorktown Heights, NY 10598, USA
/*--*/

/* System library. */

#include <sys_defs.h>

/* Utility library. */

#include <msg.h>
#include <stringops.h>
#include <stdlib.h>

/* Global library. */

#include <cleanup_user.h>
#include <mail_params.h>
#include <mail_proto.h>
#include <bounce.h>
#include <dsn_util.h>
#include <record.h>
#include <rec_type.h>
#include <dsn_smtp.h>
#include <mail_queue.h>
#include <dsn_attr_map.h>

/* Application-specific. */

#include "cleanup.h"

#define STR(x) vstring_str(x)

/* cleanup_bounce - return mail to sender */

void    cleanup_bounce(CLEANUP_STATE *state)
{
    VSTRING *buf = vstring_alloc(100);
    VSTREAM *fp;
    CLEANUP_STAT_DETAIL *detail;
    DSN_SPLIT dp;
    const char *dsn_status;
    const char *dsn_text;
    char   *rcpt = 0;
    char   *encoding;

    /*
     * Parse the failure reason if one was given, otherwise use a generic
     * mapping from cleanup-internal error code to (DSN, text).
     */
    if (state->reason) {
	dsn_split(&dp, "5.0.0", state->reason);
	dsn_status = DSN_STATUS(dp.dsn);
	dsn_text = dp.text;
    } else {
	detail = cleanup_stat_detail(state->errs);
	dsn_status = detail->dsn;
	dsn_text = detail->text;
    }

    /*
     * Create a bounce logfile with one entry for each recipient. Degrade
     * gracefully in case of no recipients or no queue file.
     */
    (void) mail_queue_path(buf, state->queue_name, state->queue_id);
    if ((fp = vstream_fopen(STR(buf), O_RDWR, 0)) != 0) {
	char   *attr_name;
	char   *attr_value;
	char   *dsn_orcpt = 0;
	int     dsn_notify = 0;
	char   *orig_rcpt = 0;
	char   *start;
	int     rec_type;
	int     junk;

	while ((state->errs & CLEANUP_STAT_WRITE) == 0
	       && (rec_type = rec_get(fp, buf, 0)) > 0
	       && rec_type != REC_TYPE_END) {
	    start = STR(buf);
	    if (rec_type == REC_TYPE_ATTR) {
		if (split_nameval(STR(buf), &attr_name, &attr_value) != 0
		    || *attr_value == 0)
		    continue;
		/* Map DSN attribute names to pseudo record type. */
		if ((junk = dsn_attr_map(attr_name)) != 0) {
		    start = attr_value;
		    rec_type = junk;
		}
	    }
	    switch (rec_type) {
	    case REC_TYPE_DSN_ORCPT:		/* RCPT TO ORCPT parameter */
		if (dsn_orcpt != 0)		/* can't happen */
		    myfree(dsn_orcpt);
		dsn_orcpt = mystrdup(start);
		break;
	    case REC_TYPE_DSN_NOTIFY:		/* RCPT TO NOTIFY parameter */
		if (alldig(start) && (junk = atoi(start)) > 0
		    && DSN_NOTIFY_OK(junk))
		    dsn_notify = junk;
		else
		    dsn_notify = 0;
		break;
	    case REC_TYPE_ORCP:		/* unmodified RCPT TO address */
		if (orig_rcpt != 0)		/* can't happen */
		    myfree(orig_rcpt);
		orig_rcpt = mystrdup(start);
		break;
	    case REC_TYPE_RCPT:		/* rewritten RCPT TO address */
		rcpt = start;
		if (bounce_append(BOUNCE_FLAG_CLEAN, state->queue_id,
				  orig_rcpt ? orig_rcpt : rcpt, rcpt,
				  (long) 0, dsn_orcpt ? dsn_orcpt : "",
				  dsn_notify, "none", dsn_status,
				  NO_DSN_DCODE, state->time,
				  "%s", dsn_text) != 0)
		    state->errs |= CLEANUP_STAT_WRITE;
		/* FALLTHROUGH */
	    case REC_TYPE_DONE:		/* can't happen */
		if (orig_rcpt != 0) {
		    myfree(orig_rcpt);
		    orig_rcpt = 0;
		}
		if (dsn_orcpt != 0) {
		    myfree(dsn_orcpt);
		    dsn_orcpt = 0;
		}
		dsn_notify = 0;
		break;
	    }
	}
	(void) vstream_fclose(fp);
	if (orig_rcpt != 0)			/* can't happen */
	    myfree(orig_rcpt);
	if (dsn_orcpt != 0)			/* can't happen */
	    myfree(dsn_orcpt);
    } else {
	msg_warn("open file %s: %m", STR(buf));
    }

    /*
     * No recipients, or no queue file.
     */
    if (rcpt == 0 && bounce_append(BOUNCE_FLAG_CLEAN, state->queue_id,
			      "unknown", "unknown", (long) 0, "", 0, "none",
				   dsn_status, NO_DSN_DCODE,
				   state->time, "%s", dsn_text) != 0)
	state->errs |= CLEANUP_STAT_WRITE;

    /*
     * Send a notification to the sender, based on the bounce logfile that we
     * just wrote.
     */
    if ((state->errs & CLEANUP_STAT_WRITE) == 0
	&& bounce_flush(BOUNCE_FLAG_CLEAN, state->queue_name,
			state->queue_id,
		(encoding = nvtable_find(state->attr, MAIL_ATTR_ENCODING)) ?
			encoding : MAIL_ATTR_ENC_NONE,
			state->sender,
			state->dsn_envid ? state->dsn_envid : "",
			state->dsn_ret) == 0) {
	state->errs = 0;
    }

    /*
     * Admit defeat and report a queue file write error so that the whole
     * operation will have to repeated at a later time.
     */
    else {
	if (var_soft_bounce == 0)
	    msg_warn("%s: bounce message failure", state->queue_id);
	state->errs |= CLEANUP_STAT_WRITE;
    }
    vstring_free(buf);
}
